/*
 * elf_endian.hpp
 *
 * Copyright (C) 2007 Bastian Blank <waldi@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef ELF_ENDIAN_HPP
#define ELF_ENDIAN_HPP

#include "elf_defs.hpp"
#include "endian.hpp"

namespace Elf
{
  template <typename from, typename type>
    struct convert
    { };

  template <typename type>
    struct convert<file_data_2LSB, type> : public endian::convert<endian::little_endian, type>
    { };

  template <typename type>
    struct convert<file_data_2MSB, type> : public endian::convert<endian::big_endian, type>
    { };
}

#endif
